import sys

from getpass import getpass

from . import check_password

count: int = check_password(getpass("Password to check: "))
if count > 0:
    print("Found password {} times!".format(count))
    sys.exit(1)
else:
    print("Could not find that password in the dataset.")



import os
import pathlib
import hashlib
import xml.etree.cElementTree as ET

from getpass import getpass
from typing import List, Optional
from xml.etree.ElementTree import Element as XMLElement

from .log import logger
from .utils import AutoRepr
from .keepass_wrapper import KeepassWrapper


class Credential(AutoRepr):
    """
    Represents one entry in a KeepassXC Database
    """

    # ordering for repr
    attrs = ["title", "username", "password", "sha1"]

    # ordering for display
    display_attrs = ["title", "username", "sha1"]

    # attributes to extract from XML
    parsed_attrs = {"title", "username", "password"}

    def __init__(self, xml_entry: XMLElement):
        self._xml_entry = xml_entry
        for str_node in filter(lambda e: e.tag == "String", list(xml_entry)):
            key_node: Optional[XMLElement] = str_node.find("Key")
            value_node: Optional[XMLElement] = str_node.find("Value")

            if key_node is not None and value_node is not None:
                key = key_node.text.lower()  # type: ignore
                value = value_node.text  # type: ignore
                if key in self.__class__.parsed_attrs:
                    setattr(self, key, value)

        if not hasattr(self, "password") or (hasattr(
                self, "password") and getattr(self, "password") is None):
            raise ValueError(
                "Ignoring entry with no password: {}".format(self))

        self._sha1: Optional[str] = None

    @property
    def sha1(self) -> Optional[str]:
        """
        Generates the sha1 from the password if needed, and returns it
        May return none, is password is none
        """
        if self._sha1 is None:
            if self.password is not None:  # type: ignore
                self._sha1 = (
                    hashlib.sha1(self.password.encode(  # type: ignore
                        "utf-8")).hexdigest().upper())
        return self._sha1

    def __eq__(self, other):
        if not isinstance(other, self.__class__):
            return False
        try:
            return self.title == other.title and self.username == other.username and self.password == other.password
        except AttributeError:
            return self.password == other.password

    def display(self):
        """An basic representation of this credential"""
        # must have at least a password, which would be displayed as sha1
        for a in self.__class__.display_attrs:
            try:
                d = getattr(self, a)
                if d is not None:
                    return d
            except AttributeError:
                pass


class Database(AutoRepr):

    attrs = ["database_file", "key_file"]

    def __init__(
        self,
        database_file: pathlib.Path,
        key_file: Optional[pathlib.Path] = None,
        keepassxc_cli_location: Optional[pathlib.Path] = None,
    ):
        self.database_file = database_file
        self.key_file = key_file

        self._xml_tree: Optional[XMLElement] = None
        self._password: Optional[str] = None
        self._credentials: Optional[List[Credential]] = None

        self.keepass_wrapper = KeepassWrapper(keepassxc_cli_location)

    @property
    def password(self) -> str:
        """
        Returns the password for this database
        If the KEEPASSXC_PWNED_PASSWD environment variable is set,
        uses that, else, prompts the user for password
        """
        if self._password is not None:
            return self._password
        elif "KEEPASSXC_PWNED_PASSWD" in os.environ:
            logger.debug(
                "Using password from KEEPASSXC_PWNED_PASSWD environment variable"
            )
            self._password = os.environ["KEEPASSXC_PWNED_PASSWD"]
            return self._password
        else:
            self._password = getpass("Insert password for {}: ".format(
                self.database_file))
            return self._password

    @property
    def xml_tree(self) -> Optional[XMLElement]:
        """
        Returns the parsed XML Element Tree from the keepassxc-cli export command
        Calls the command if it hasn't been called yet
        """
        self._call_keepassxc_cli()
        return self._xml_tree

    def _call_keepassxc_cli(self) -> None:
        """
        Calls the keepassxc-cli export as a subprocess
        Sets the '_xml_tree' attribute to the XML representation of the database
        """
        if self._xml_tree is not None:
            return  # already called, use cached value
        keepass_export_process_output: str = self.keepass_wrapper.export_database(
            database_file=self.database_file,
            database_password=self.password,  # calls getpass if not set
            database_keyfile=self.key_file,
        )
        self._xml_tree = ET.fromstring(keepass_export_process_output)

    @property
    def credentials(self) -> List[Credential]:
        """
        Returns a list of credentials -- entries from the KDBX
        """
        if self._credentials is not None:
            return self._credentials
        self._credentials = []
        for group in self.xml_tree.find("Root").iter("Group"):  # type: ignore
            # ignore deleted passwords
            if group.find("Name").text == "Recycle Bin":  # type: ignore
                continue
            # grab username, title, and passwords
            for entry in filter(lambda g: g.tag == "Entry", list(group)):
                try:
                    cred = Credential(entry)
                    self._credentials.append(cred)
                except ValueError as no_pw:
                    logger.debug(str(no_pw))
        logger.debug("KeepassXC parsed entry count: {}".format(
            len(self._credentials)))
        return self._credentials



import pathlib
import logging

from typing import Mapping, List

import click

from .log import logger, set_level
from .exceptions import PwnedPasswordException


@click.command()
@click.option(
    "-p",
    "--plaintext",
    default=False,
    is_flag=True,
    help="Print breached passwords in plaintext; defaults to sha1 hashes.")
@click.option("-k",
              "--key-file",
              required=False,
              type=click.Path(exists=True),
              help="Key file for the database")
@click.option("-v",
              "--verbose",
              default=False,
              is_flag=True,
              help="Print debug messages")
@click.option("-q",
              "--quiet",
              default=False,
              is_flag=True,
              help="Don't print status messages, just the summary")
@click.option("--keepassxc-cli",
              type=click.Path(exists=True),
              help="Specify a different location for the keepassxc-cli binary")
@click.argument("database", required=True, type=click.Path(exists=True))
def main(plaintext, key_file, verbose, quiet, database, keepassxc_cli):
    """Check a keepassxc database against previously cracked haveibeenpwned passwords"""
    main_wrapper(plaintext, key_file, verbose, quiet, database, keepassxc_cli)


def main_wrapper(plaintext, key_file, verbose, quiet, database,
                 keepassxc_cli_location):
    """Called from main click command"""

    # setup logs before other imports to ensure correct log level
    log_level: int = logging.INFO
    # cant run verbose and quiet simultaneously, choose higher value
    if verbose:
        log_level = logging.DEBUG
    elif quiet:
        log_level = logging.ERROR

    set_level(log_level)

    from .parser import Database, Credential
    from .cache import PasswordCache

    if keepassxc_cli_location is not None:
        keepassxc_cli_location = pathlib.Path(keepassxc_cli_location)
    if key_file is not None:
        key_file: pathlib.Path = pathlib.Path(key_file)
    db: Database = Database(pathlib.Path(database), key_file,
                            keepassxc_cli_location)
    # maps sha1 hashes to credentials and their occurrence counts
    breached_passwords: List[Credential] = []
    pw_cache: Mapping[str, int] = PasswordCache()
    for credential in db.credentials:
        logger.info("Checking password for {}...".format(credential.display()))
        # pw_cache __missing__ makes the http request to get the count
        try:
            occurrence_count = pw_cache[credential.sha1]
            if occurrence_count > 0:
                logger.info(
                    "Found password for '{}' {} times in the dataset!".format(
                        credential.display(), occurrence_count))
                breached_passwords.append(credential)
        except PwnedPasswordException as http_err:
            logger.critical(str(http_err))
            logger.critical("Ignoring previous entry due to HTTP error")
    breached_passwords_count = len(breached_passwords)
    if breached_passwords_count > 0:
        print("Found {} previously breached password{}:".format(
            breached_passwords_count,
            "s" if breached_passwords_count > 1 else ""))
        for credential in breached_passwords:
            display_pw: str = credential.password if plaintext else credential.sha1
            print("{}:{}:{}".format(credential.display(), display_pw,
                                    pw_cache[credential.sha1]))
    else:
        print("None of your passwords have been found breached.")



class AutoRepr:
    """
    Automatically creates a repr for a class given a list of attributes
    """

    # must have self.__class__.attrs set
    def __repr__(self) -> str:
        assert self.__class__.attrs is not None  # type: ignore

        return "{}({})".format(
            self.__class__.__name__,
            ", ".join([
                "{}={}".format(a, repr(getattr(self, a, None)))
                for a in self.__class__.attrs  # type: ignore
            ]),
        )

    def __str__(self) -> str:
        return self.__repr__()



from typing import Mapping, Any

from .utils import AutoRepr


class PwnedPasswordException(AutoRepr, Exception):
    """
    PwnedPassword API Client Errors (>=400, excluding 429)
    """

    attrs = ["url", "status_code", "http_text"]

    def __init__(self, message: Mapping[str, Any]):
        self.url = message["url"]
        self.status_code = message["status_code"]
        self.http_text = message["http_text"]

        super(PwnedPasswordException, self).__init__(str(self))

    def __str__(self) -> str:
        return "Error fetching {}\n{}: {}".format(self.url, self.status_code,
                                                  self.http_text)



import sys
import logging

logger: logging.Logger = logging.getLogger("keepassxc_pwned")


def setup_logs(log_level: int = logging.INFO):
    """
    Setup the global logger
    This does not setup logs more than once.
    """
    global logger
    if logger.handlers:
        return  # logs have already been set up
    logger = logging.getLogger("keepassxc_pwned")
    sh = logging.StreamHandler(sys.stdout)
    sh.setLevel(log_level)
    sh.setFormatter(logging.Formatter("%(msg)s"))
    logger.addHandler(sh)


def set_level(log_level: int = logging.INFO):
    """
    Sets the log level for the application
    """
    global logger
    logger.setLevel(log_level)
    for handler in logger.handlers:
        handler.setLevel(log_level)



import time
import hashlib

from typing import Optional

import requests

from .log import logger
from .exceptions import PwnedPasswordException

default_headers = {"User-Agent": "keepassxc-pwned"}


def request_password_hash(hash_head: str) -> requests.Response:
    """
    Requests the Response object for the corresponding hash head (5 chars)
    Raises PwnedPasswordException on unrecoverable errors
    """
    url = "https://api.pwnedpasswords.com/range/" + hash_head
    logger.debug("Requesting {}".format(url))
    res = requests.get(url, headers=default_headers)
    if res.status_code >= 400:
        # on occasion I've had random 409 cloudflare errors
        if res.status_code in [
                400, 403, 404
        ]:  # https://haveibeenpwned.com/API/v2#ResponseCodes
            raise PwnedPasswordException({
                "url": url,
                "status_code": res.status_code,
                "http_text": res.text
            })
        else:
            logger.warning("Request failed, retrying...")
            time.sleep(10)
            return request_password_hash(hash_head)
    return res


# Adapted from: https://github.com/mikepound/pwned-search/blob/8efd8ffedd398756e26d52ef51206ba6d8e28f57/pwned.py#L12
def lookup_pwned(passwd: Optional[str] = None,
                 pw_hash: Optional[str] = None) -> int:
    """
    Returns number of times password was seen in pwned database
    Raises PwnedPasswordException on unrecoverable errors
    """
    if pw_hash is None:
        if passwd is not None:
            pw_hash = hashlib.sha1(passwd.encode("utf-8")).hexdigest().upper()
        else:
            raise ValueError("You must pass either 'passwd' or 'pw_hash'")
    head, tail = pw_hash[:5], pw_hash[5:]
    res: requests.Response = request_password_hash(head)
    hashes = (line.split(":") for line in res.text.splitlines())
    count = next((int(count) for t, count in hashes if t == tail), 0)
    return count



import sys
import os
import shlex
import subprocess
import pathlib
import shutil

from typing import Optional, List
from functools import partial
from distutils.version import StrictVersion

from .log import logger

# redirect STDIN/STDERR from subprocess
subprocess_piped = partial(subprocess.run,
                           encoding="utf-8",
                           stdout=subprocess.PIPE,
                           stderr=subprocess.PIPE)


class KeepassWrapper:
    """
    Functions that create a subprocess and call the keepassxc-cli shell command
    """

    KEEPASSXC_CLI_NAME = "keepassxc-cli"

    def __init__(self, keepassxc_cli_location: Optional[pathlib.Path] = None):
        """
        Create an instance of KeepassWrapper, with an optional
        path to the keepassxc-cli binary.


        Allow the path of the keepassxc-cli to be changed because
        some package managers (e.g. mint with snap) call
        this the binary keepassxc.cli
        """

        if keepassxc_cli_location is None:
            # search for keepassxc-cli which default name on $PATH
            keepassxc_location_output = shutil.which(
                self.__class__.KEEPASSXC_CLI_NAME)
            if keepassxc_location_output is None:
                logger.critical(
                    "Could not find a binary called {} on your $PATH.".format(
                        self.__class__.KEEPASSXC_CLI_NAME))
                sys.exit(1)
            self.keepassxc_cli_location: pathlib.Path = pathlib.Path(
                keepassxc_location_output)
        else:
            self.keepassxc_cli_location = keepassxc_cli_location
        logger.debug("keepassxc-cli location: {}".format(
            self.keepassxc_cli_location))

    def version(self) -> StrictVersion:
        """Returns the KeepassXC Version"""
        version_proc: subprocess.CompletedProcess = subprocess_piped(
            [self.keepassxc_cli_location, "--version"])
        version_str: str = version_proc.stdout.strip()
        logger.debug("keepassxc cli version: {}".format(version_str))
        return StrictVersion(version_str)

    def backwards_compatible_export(self) -> str:
        """
        In KeepassXC version 2.5.0, the extract command was re-named to export
        Attempt to parse the version number and generate the correct subcommand
        """
        try:
            version: StrictVersion = self.version()
            if version < StrictVersion("2.5.0"):
                return "extract"
            else:
                return "export"
        except ValueError:
            return "export"

    def export_database(
        self,
        database_file: pathlib.Path,
        database_password: str,
        database_keyfile: Optional[pathlib.Path] = None,
        keepassxc_cli_location: Optional[pathlib.Path] = None,
    ) -> str:
        """Calls the keepassxc-cli export command, returns the output from the command"""

        command_parts: List[str] = [
            str(self.keepassxc_cli_location),
            self.backwards_compatible_export()
        ]
        if database_keyfile is not None:
            command_parts.extend(["-k", str(database_keyfile)])
        command_parts.append(str(database_file))
        command_str: str = " ".join(command_parts)
        logger.debug("Export database command: {}".format(command_str))
        keepassxc_output: subprocess.CompletedProcess = subprocess_piped(
            shlex.split(command_str),
            input=database_password,
        )
        if keepassxc_output.returncode != 0:
            logger.critical("Error communicating with {}".format(
                self.keepassxc_cli_location))
            logger.critical(keepassxc_output.stderr)
            sys.exit(1)
        # python doesn't like the version tag, remove the first line
        return keepassxc_output.stdout.split(os.linesep, 2)[-1]



from time import sleep

from .password import lookup_pwned


class PasswordCache(dict):
    """
    If the password hash has not been requested before, requests it.
    If you have the same password for many entries, this would only request that once.
    """
    def __missing__(self, key: str) -> int:
        """
        Looks up the occurrence count for the sha1 value passed as key
        """
        count: int = lookup_pwned(pw_hash=key)
        sleep(2)
        self[key] = count
        return count

    def __setitem__(self, key: str, value: int):
        """
        sets sha1 -> occurence_count
        """
        super(PasswordCache, self).__setitem__(key, value)



#!/usr/bin/env python
# -*- coding: utf-8 -*-

import io
from setuptools import setup, find_packages

with io.open("requirements.txt", 'r', encoding='utf-8') as req_file:
    requirements = req_file.read().splitlines()

with io.open("README.md", 'r', encoding='utf-8') as readme:
    readme_contents = readme.read()

setup(
    author="Sean Breckenridge",
    author_email='seanbrecke@gmail.com',
    classifiers=[
        'Development Status :: 3 - Alpha',
        'Intended Audience :: End Users/Desktop',
        'Natural Language :: English',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7',
        'Programming Language :: Python :: 3.8',
        'Programming Language :: Python :: 3.9',
    ],
    install_requires=requirements,
    description="Check your keepassxc database against previously breached haveibeenpwned passwords",
    license="MIT",
    long_description=readme_contents,
    long_description_content_type="text/markdown",
    include_package_data=True,
    keywords='keepassxc password-strength password',
    name='keepassxc-pwned',
    packages=find_packages(include=['keepassxc_pwned']),
    entry_points = {
        'console_scripts': [
            "keepassxc_pwned = keepassxc_pwned.cli:main"
        ]
    },
    url='https://github.com/seanbreckenridge/keepassxc-pwned',
    version='0.3.1',
    zip_safe=True,
)
